@@ -1,32 +1,42 @@
-# -*- coding: utf-8 -*-
-from collections import defaultdict
-from concurrent.futures import ThreadPoolExecutor
-from datetime import datetime
-import io
+"""Core functions."""
+
+from __future__ import annotations
+
 import itertools
 import logging
 import operator
-import os.path
-import socket
+import os
+from collections import defaultdict
+from concurrent.futures import ThreadPoolExecutor
+from datetime import datetime, timedelta, timezone
+from typing import TYPE_CHECKING, Any
+from zipfile import BadZipfile
 
-from babelfish import Language, LanguageReverseError
-from guessit import guessit
-from rarfile import NotRarFile, RarCannotExec, RarFile
-import requests
+from babelfish import Language, LanguageReverseError  # type: ignore[import-untyped]
+from guessit import guessit  # type: ignore[import-untyped]
+from rarfile import BadRarFile, Error, NotRarFile, RarCannotExec, RarFile, is_rarfile  # type: ignore[import-untyped]
 
-from .extensions import provider_manager, refiner_manager
+from .extensions import default_providers, provider_manager, refiner_manager
 from .score import compute_score as default_compute_score
-from .subtitle import SUBTITLE_EXTENSIONS, get_subtitle_path
-from .utils import hash_napiprojekt, hash_opensubtitles, hash_shooter, hash_thesubdb
+from .subtitle import SUBTITLE_EXTENSIONS, Subtitle
+from .utils import handle_exception
 from .video import VIDEO_EXTENSIONS, Episode, Movie, Video
 
+if TYPE_CHECKING:
+    from collections.abc import Iterator, Mapping, Sequence, Set
+    from types import TracebackType
+
+    from subliminal.providers import Provider
+    from subliminal.score import ComputeScore
+
+
 #: Supported archive extensions
 ARCHIVE_EXTENSIONS = ('.rar',)
 
 logger = logging.getLogger(__name__)
 
 
-class ProviderPool(object):
+class ProviderPool:
     """A pool of providers with the same API as a single :class:`~subliminal.providers.Provider`.
 
     It has a few extra features:
@@ -37,31 +47,44 @@ class ProviderPool(object):
 
     :param list providers: name of providers to use, if not all.
     :param dict provider_configs: provider configuration as keyword arguments per provider name to pass when
-        instanciating the :class:`~subliminal.providers.Provider`.
+        instantiating the :class:`~subliminal.providers.Provider`.
 
     """
-    def __init__(self, providers=None, provider_configs=None):
-        #: Name of providers to use
-        self.providers = providers or provider_manager.names()
 
-        #: Provider configuration
+    #: Name of providers to use
+    providers: Sequence[str]
+    #: Provider configuration
+    provider_configs: Mapping[str, Any]
+    #: Initialized providers
+    initialized_providers: dict[str, Provider]
+    #: Discarded providers
+    discarded_providers: set[str]
+
+    def __init__(
+        self,
+        providers: Sequence[str] | None = None,
+        provider_configs: Mapping[str, Any] | None = None,
+    ) -> None:
+        self.providers = providers or default_providers
         self.provider_configs = provider_configs or {}
-
-        #: Initialized providers
         self.initialized_providers = {}
-
-        #: Discarded providers
         self.discarded_providers = set()
 
-    def __enter__(self):
+    def __enter__(self) -> ProviderPool:
         return self
 
-    def __exit__(self, exc_type, exc_value, traceback):
+    def __exit__(
+        self,
+        exc_type: type[BaseException],
+        exc_value: BaseException | None,
+        traceback: TracebackType | None,
+    ) -> None:
         self.terminate()
 
-    def __getitem__(self, name):
+    def __getitem__(self, name: str) -> Provider:
         if name not in self.providers:
             raise KeyError
+
         if name not in self.initialized_providers:
             logger.info('Initializing provider %s', name)
             provider = provider_manager[name].plugin(**self.provider_configs.get(name, {}))
@@ -70,24 +93,22 @@ class ProviderPool(object):
 
         return self.initialized_providers[name]
 
-    def __delitem__(self, name):
+    def __delitem__(self, name: str) -> None:
         if name not in self.initialized_providers:
             raise KeyError(name)
 
         try:
             logger.info('Terminating provider %s', name)
             self.initialized_providers[name].terminate()
-        except (requests.Timeout, socket.timeout):
-            logger.error('Provider %r timed out, improperly terminated', name)
-        except:
-            logger.exception('Provider %r terminated unexpectedly', name)
+        except Exception as e:  # noqa: BLE001
+            handle_exception(e, f'Provider {name} improperly terminated')
 
         del self.initialized_providers[name]
 
-    def __iter__(self):
+    def __iter__(self) -> Iterator[str]:
         return iter(self.initialized_providers)
 
-    def list_subtitles_provider(self, provider, video, languages):
+    def list_subtitles_provider(self, provider: str, video: Video, languages: Set[Language]) -> list[Subtitle]:
         """List subtitles with a single provider.
 
         The video and languages are checked against the provider.
@@ -107,7 +128,7 @@ class ProviderPool(object):
             return []
 
         # check supported languages
-        provider_languages = provider_manager[provider].plugin.languages & languages
+        provider_languages = provider_manager[provider].plugin.check_languages(languages)
         if not provider_languages:
             logger.info('Skipping provider %r: no language to search for', provider)
             return []
@@ -116,12 +137,12 @@ class ProviderPool(object):
         logger.info('Listing subtitles with provider %r and languages %r', provider, provider_languages)
         try:
             return self[provider].list_subtitles(video, provider_languages)
-        except (requests.Timeout, socket.timeout):
-            logger.error('Provider %r timed out', provider)
-        except:
-            logger.exception('Unexpected error in provider %r', provider)
+        except Exception as e:  # noqa: BLE001
+            handle_exception(e, 'Provider {provider}')
 
-    def list_subtitles(self, video, languages):
+        return []
+
+    def list_subtitles(self, video: Video, languages: Set[Language]) -> list[Subtitle]:
         """List subtitles.
 
         :param video: video to list subtitles for.
@@ -152,7 +173,7 @@ class ProviderPool(object):
 
         return subtitles
 
-    def download_subtitle(self, subtitle):
+    def download_subtitle(self, subtitle: Subtitle) -> bool:
         """Download `subtitle`'s :attr:`~subliminal.subtitle.Subtitle.content`.
 
         :param subtitle: subtitle to download.
@@ -169,14 +190,11 @@ class ProviderPool(object):
         logger.info('Downloading subtitle %r', subtitle)
         try:
             self[subtitle.provider_name].download_subtitle(subtitle)
-        except (requests.Timeout, socket.timeout):
-            logger.error('Provider %r timed out, discarding it', subtitle.provider_name)
+        except (BadZipfile, BadRarFile):
+            logger.exception('Bad archive for subtitle %r', subtitle)
+        except Exception as e:  # noqa: BLE001
+            handle_exception(e, f'Discarding provider {subtitle.provider_name}')
             self.discarded_providers.add(subtitle.provider_name)
-            return False
-        except:
-            logger.exception('Unexpected error in provider %r, discarding it', subtitle.provider_name)
-            self.discarded_providers.add(subtitle.provider_name)
-            return False
 
         # check subtitle validity
         if not subtitle.is_valid():
@@ -185,8 +203,17 @@ class ProviderPool(object):
 
         return True
 
-    def download_best_subtitles(self, subtitles, video, languages, min_score=0, hearing_impaired=False, only_one=False,
-                                compute_score=None):
+    def download_best_subtitles(
+        self,
+        subtitles: Sequence[Subtitle],
+        video: Video,
+        languages: Set[Language],
+        *,
+        min_score: int = 0,
+        hearing_impaired: bool = False,
+        only_one: bool = False,
+        compute_score: ComputeScore | None = None,
+    ) -> list[Subtitle]:
         """Download the best matching subtitles.
 
         :param subtitles: the subtitles to use.
@@ -207,11 +234,14 @@ class ProviderPool(object):
         compute_score = compute_score or default_compute_score
 
         # sort subtitles by score
-        scored_subtitles = sorted([(s, compute_score(s, video, hearing_impaired=hearing_impaired))
-                                  for s in subtitles], key=operator.itemgetter(1), reverse=True)
+        scored_subtitles = sorted(
+            [(s, compute_score(s, video, hearing_impaired=hearing_impaired)) for s in subtitles],
+            key=operator.itemgetter(1),
+            reverse=True,
+        )
 
         # download best subtitles, falling back on the next on error
-        downloaded_subtitles = []
+        downloaded_subtitles: list[Subtitle] = []
         for subtitle, score in scored_subtitles:
             # check score
             if score < min_score:
@@ -219,7 +249,7 @@ class ProviderPool(object):
                 break
 
             # check downloaded languages
-            if subtitle.language in set(s.language for s in downloaded_subtitles):
+            if subtitle.language in {s.language for s in downloaded_subtitles}:
                 logger.debug('Skipping subtitle: %r already downloaded', subtitle.language)
                 continue
 
@@ -228,18 +258,18 @@ class ProviderPool(object):
                 downloaded_subtitles.append(subtitle)
 
             # stop when all languages are downloaded
-            if set(s.language for s in downloaded_subtitles) == languages:
+            if {s.language for s in downloaded_subtitles} == languages:
                 logger.debug('All languages downloaded')
                 break
 
             # stop if only one subtitle is requested
-            if only_one:
+            if only_one and len(downloaded_subtitles) > 0:
                 logger.debug('Only one subtitle downloaded')
                 break
 
         return downloaded_subtitles
 
-    def terminate(self):
+    def terminate(self) -> None:
         """Terminate all the :attr:`initialized_providers`."""
         logger.debug('Terminating initialized providers')
         for name in list(self.initialized_providers):
@@ -253,22 +283,36 @@ class AsyncProviderPool(ProviderPool):
         to the number of :attr:`~ProviderPool.providers`.
 
     """
-    def __init__(self, max_workers=None, *args, **kwargs):
-        super(AsyncProviderPool, self).__init__(*args, **kwargs)
+
+    max_workers: int
+
+    def __init__(self, max_workers: int | None = None, *args: Any, **kwargs: Any) -> None:
+        super().__init__(*args, **kwargs)
 
         #: Maximum number of threads to use
         self.max_workers = max_workers or len(self.providers)
 
-    def list_subtitles_provider(self, provider, video, languages):
-        return provider, super(AsyncProviderPool, self).list_subtitles_provider(provider, video, languages)
+    def list_subtitles_provider_tuple(
+        self,
+        provider: str,
+        video: Video,
+        languages: Set[Language],
+    ) -> tuple[str, list[Subtitle]]:
+        """List subtitles with a single provider, multi-threaded."""
+        return provider, super().list_subtitles_provider(provider, video, languages)
 
-    def list_subtitles(self, video, languages):
+    def list_subtitles(self, video: Video, languages: Set[Language]) -> list[Subtitle]:
+        """List subtitles, multi-threaded."""
         subtitles = []
 
         with ThreadPoolExecutor(self.max_workers) as executor:
-            for provider, provider_subtitles in executor.map(self.list_subtitles_provider, self.providers,
-                                                             itertools.repeat(video, len(self.providers)),
-                                                             itertools.repeat(languages, len(self.providers))):
+            executor_map = executor.map(
+                self.list_subtitles_provider_tuple,
+                self.providers,
+                itertools.repeat(video, len(self.providers)),
+                itertools.repeat(languages, len(self.providers)),
+            )
+            for provider, provider_subtitles in executor_map:
                 # discard provider that failed
                 if provider_subtitles is None:
                     logger.info('Discarding provider %s', provider)
@@ -281,7 +325,13 @@ class AsyncProviderPool(ProviderPool):
         return subtitles
 
 
-def check_video(video, languages=None, age=None, undefined=False):
+def check_video(
+    video: Video,
+    *,
+    languages: Set[Language] | None = None,
+    age: timedelta | None = None,
+    undefined: bool = False,
+) -> bool:
     """Perform some checks on the `video`.
 
     All the checks are optional. Return `False` if any of this check fails:
@@ -318,7 +368,11 @@ def check_video(video, languages=None, age=None, undefined=False):
     return True
 
 
-def search_external_subtitles(path, directory=None):
+def search_external_subtitles(
+    path: str | os.PathLike,
+    *,
+    directory: str | os.PathLike | None = None,
+) -> dict[str, Language]:
     """Search for external subtitles from a video `path` and their associated language.
 
     Unless `directory` is provided, search will be made in the same directory as the video file.
@@ -338,17 +392,17 @@ def search_external_subtitles(path, directory=None):
     subtitles = {}
     for p in os.listdir(directory or dirpath):
         # keep only valid subtitle filenames
-        if not p.startswith(fileroot) or not p.endswith(SUBTITLE_EXTENSIONS):
+        if not p.startswith(fileroot) or not p.lower().endswith(SUBTITLE_EXTENSIONS):
             continue
 
         # extract the potential language code
         language = Language('und')
-        language_code = p[len(fileroot):-len(os.path.splitext(p)[1])].replace(fileext, '').replace('_', '-')[1:]
+        language_code = p[len(fileroot) : -len(os.path.splitext(p)[1])].replace(fileext, '').replace('_', '-')[1:]
         if language_code:
             try:
                 language = Language.fromietf(language_code)
             except (ValueError, LanguageReverseError):
-                logger.error('Cannot parse language code %r', language_code)
+                logger.exception('Cannot parse language code %r', language_code)
 
         subtitles[p] = language
 
@@ -357,21 +411,24 @@ def search_external_subtitles(path, directory=None):
     return subtitles
 
 
-def scan_video(path):
+def scan_video(path: str | os.PathLike) -> Video:
     """Scan a video from a `path`.
 
     :param str path: existing path to the video.
     :return: the scanned video.
     :rtype: :class:`~subliminal.video.Video`
-
+    :raises: :class:`ValueError`: video path is not well defined.
     """
+    path = os.fspath(path)
     # check for non-existing path
     if not os.path.exists(path):
-        raise ValueError('Path does not exist')
+        msg = 'Path does not exist'
+        raise ValueError(msg)
 
     # check video extension
-    if not path.endswith(VIDEO_EXTENSIONS):
-        raise ValueError('%r is not a valid video extension' % os.path.splitext(path)[1])
+    if not path.lower().endswith(VIDEO_EXTENSIONS):
+        msg = f'{os.path.splitext(path)[1]!r} is not a valid video extension'
+        raise ValueError(msg)
 
     dirpath, filename = os.path.split(path)
     logger.info('Scanning video %r in %r', filename, dirpath)
@@ -379,69 +436,72 @@ def scan_video(path):
     # guess
     video = Video.fromguess(path, guessit(path))
 
-    # size and hashes
+    # size
     video.size = os.path.getsize(path)
-    if video.size > 10485760:
-        logger.debug('Size is %d', video.size)
-        video.hashes['opensubtitles'] = hash_opensubtitles(path)
-        video.hashes['shooter'] = hash_shooter(path)
-        video.hashes['thesubdb'] = hash_thesubdb(path)
-        video.hashes['napiprojekt'] = hash_napiprojekt(path)
-        logger.debug('Computed hashes %r', video.hashes)
-    else:
-        logger.warning('Size is lower than 10MB: hashes not computed')
+    logger.debug('Size is %d', video.size)
 
     return video
 
 
-def scan_archive(path):
+def scan_archive(path: str | os.PathLike) -> Video:
     """Scan an archive from a `path`.
 
     :param str path: existing path to the archive.
     :return: the scanned video.
     :rtype: :class:`~subliminal.video.Video`
-
+    :raises: :class:`ValueError`: video path is not well defined.
     """
+    path = os.fspath(path)
     # check for non-existing path
     if not os.path.exists(path):
-        raise ValueError('Path does not exist')
+        msg = 'Path does not exist'
+        raise ValueError(msg)
 
-    # check video extension
-    if not path.endswith(ARCHIVE_EXTENSIONS):
-        raise ValueError('%r is not a valid archive extension' % os.path.splitext(path)[1])
+    if not is_rarfile(path):
+        msg = f'{os.path.splitext(path)[1]!r} is not a valid archive'
+        raise ValueError(msg)
 
-    dirpath, filename = os.path.split(path)
-    logger.info('Scanning archive %r in %r', filename, dirpath)
+    dir_path, filename = os.path.split(path)
 
-    # rar extension
-    if filename.endswith('.rar'):
-        rar = RarFile(path)
+    logger.info('Scanning archive %r in %r', filename, dir_path)
 
-        # filter on video extensions
-        rar_filenames = [f for f in rar.namelist() if f.endswith(VIDEO_EXTENSIONS)]
+    # Get filename and file size from RAR
+    rar = RarFile(path)
 
-        # no video found
-        if not rar_filenames:
-            raise ValueError('No video in archive')
+    # check that the rar doesnt need a password
+    if rar.needs_password():
+        msg = 'Rar requires a password'
+        raise ValueError(msg)
 
-        # more than one video found
-        if len(rar_filenames) > 1:
-            raise ValueError('More than one video in archive')
+    # raise an exception if the rar file is broken
+    # must be called to avoid a potential deadlock with some broken rars
+    rar.testrar()
 
-        # guess
-        rar_filename = rar_filenames[0]
-        rar_filepath = os.path.join(dirpath, rar_filename)
-        video = Video.fromguess(rar_filepath, guessit(rar_filepath))
+    file_infos = [f for f in rar.infolist() if not f.isdir() and f.filename.endswith(VIDEO_EXTENSIONS)]
 
-        # size
-        video.size = rar.getinfo(rar_filename).file_size
-    else:
-        raise ValueError('Unsupported extension %r' % os.path.splitext(path)[1])
+    # sort by file size descending, the largest video in the archive is the one we want, there may be samples or intros
+    file_infos.sort(key=operator.attrgetter('file_size'), reverse=True)
+
+    # no video found
+    if not file_infos:
+        msg = 'No video in archive'
+        raise ValueError(msg)
+
+    # Free the information about irrelevant files before guessing
+    file_info = file_infos[0]
+
+    # guess
+    video_filename = file_info.filename
+    video_path = os.path.join(dir_path, video_filename)
+    video = Video.fromguess(video_path, guessit(video_path))
+
+    # size
+    video.size = file_info.file_size
 
     return video
 
 
-def scan_videos(path, age=None, archives=True):
+def scan_videos(path: str | os.PathLike, *, age: timedelta | None = None, archives: bool = True) -> list[Video]:
     """Scan `path` for videos and their subtitles.
 
     See :func:`refine` to find additional information for the video.
@@ -451,15 +511,18 @@ def scan_videos(path, age=None, archives=True):
     :param bool archives: scan videos in archives.
     :return: the scanned videos.
     :rtype: list of :class:`~subliminal.video.Video`
-
+    :raises: :class:`ValueError`: video path is not well defined.
     """
+    path = os.fspath(path)
     # check for non-existing path
     if not os.path.exists(path):
-        raise ValueError('Path does not exist')
+        msg = 'Path does not exist'
+        raise ValueError(msg)
 
     # check for non-directory path
     if not os.path.isdir(path):
-        raise ValueError('Path is not a directory')
+        msg = 'Path is not a directory'
+        raise ValueError(msg)
 
     # walk the path
     videos = []
@@ -471,17 +534,27 @@ def scan_videos(path, age=None, archives=True):
             if dirname.startswith('.'):
                 logger.debug('Skipping hidden dirname %r in %r', dirname, dirpath)
                 dirnames.remove(dirname)
+            # Skip Sample folder
+            if dirname.lower() == 'sample':
+                logger.debug('Skipping sample dirname %r in %r', dirname, dirpath)
+                dirnames.remove(dirname)
 
         # scan for videos
         for filename in filenames:
             # filter on videos and archives
-            if not (filename.endswith(VIDEO_EXTENSIONS) or archives and filename.endswith(ARCHIVE_EXTENSIONS)):
+            if not filename.lower().endswith(VIDEO_EXTENSIONS) and not (
+                archives and filename.lower().endswith(ARCHIVE_EXTENSIONS)
+            ):
                 continue
 
             # skip hidden files
             if filename.startswith('.'):
                 logger.debug('Skipping hidden filename %r in %r', filename, dirpath)
                 continue
+            # skip 'sample' media files
+            if os.path.splitext(filename)[0].lower() == 'sample':
+                logger.debug('Skipping sample filename %r in %r', filename, dirpath)
+                continue
 
             # reconstruct the file path
             filepath = os.path.join(dirpath, filename)
@@ -492,32 +565,46 @@ def scan_videos(path, age=None, archives=True):
                 continue
 
             # skip old files
-            if age and datetime.utcnow() - datetime.utcfromtimestamp(os.path.getmtime(filepath)) > age:
-                logger.debug('Skipping old file %r in %r', filename, dirpath)
+            try:
+                file_age = datetime.fromtimestamp(os.path.getmtime(filepath), timezone.utc)
+            except ValueError:
+                logger.warning('Could not get age of file %r in %r', filename, dirpath)
                 continue
+            else:
+                if age and datetime.now(timezone.utc) - file_age > age:
+                    logger.debug('Skipping old file %r in %r', filename, dirpath)
+                    continue
 
             # scan
-            if filename.endswith(VIDEO_EXTENSIONS):  # video
+            if filename.lower().endswith(VIDEO_EXTENSIONS):  # video
                 try:
                     video = scan_video(filepath)
                 except ValueError:  # pragma: no cover
                     logger.exception('Error scanning video')
                     continue
-            elif archives and filename.endswith(ARCHIVE_EXTENSIONS):  # archive
+            elif archives and filename.lower().endswith(ARCHIVE_EXTENSIONS):  # archive
                 try:
                     video = scan_archive(filepath)
-                except (NotRarFile, RarCannotExec, ValueError):  # pragma: no cover
+                except (Error, NotRarFile, RarCannotExec, ValueError):  # pragma: no cover
                     logger.exception('Error scanning archive')
                     continue
             else:  # pragma: no cover
-                raise ValueError('Unsupported file %r' % filename)
+                msg = f'Unsupported file {filename!r}'
+                raise ValueError(msg)
 
             videos.append(video)
 
     return videos
 
 
-def refine(video, episode_refiners=None, movie_refiners=None, **kwargs):
+def refine(
+    video: Video,
+    *,
+    episode_refiners: Sequence[str] | None = None,
+    movie_refiners: Sequence[str] | None = None,
+    refiner_configs: Mapping[str, Any] | None = None,
+    **kwargs: Any,
+) -> Video:
     """Refine a video using :ref:`refiners`.
 
     .. note::
@@ -528,23 +615,33 @@ def refine(video, episode_refiners=None, movie_refiners=None, **kwargs):
     :type video: :class:`~subliminal.video.Video`
     :param tuple episode_refiners: refiners to use for episodes.
     :param tuple movie_refiners: refiners to use for movies.
-    :param \*\*kwargs: additional parameters for the :func:`~subliminal.refiners.refine` functions.
+    :param dict refiner_configs: refiner configuration as keyword arguments per refiner name to pass when
+        calling the refine method
+    :param kwargs: additional parameters for the :func:`~subliminal.refiners.refine` functions.
 
     """
-    refiners = ()
+    refiners: tuple[str, ...] = ()
     if isinstance(video, Episode):
-        refiners = episode_refiners or ('metadata', 'tvdb', 'omdb')
+        refiners = tuple(episode_refiners) if episode_refiners is not None else ('metadata', 'tvdb', 'omdb', 'tmdb')
     elif isinstance(video, Movie):
-        refiners = movie_refiners or ('metadata', 'omdb')
-    for refiner in refiners:
+        refiners = tuple(movie_refiners) if movie_refiners is not None else ('metadata', 'omdb', 'tmdb')
+
+    for refiner in ('hash', *refiners):
         logger.info('Refining video with %s', refiner)
         try:
-            refiner_manager[refiner].plugin(video, **kwargs)
-        except:
-            logger.exception('Failed to refine video')
+            refiner_manager[refiner].plugin(video, **dict((refiner_configs or {}).get(refiner, {}), **kwargs))
+        except Exception as e:  # noqa: BLE001
+            handle_exception(e, f'Failed to refine video {video.name!r}')
+    return video
 
 
-def list_subtitles(videos, languages, pool_class=ProviderPool, **kwargs):
+def list_subtitles(
+    videos: Set[Video],
+    languages: Set[Language],
+    *,
+    pool_class: type[ProviderPool] = ProviderPool,
+    **kwargs: Any,
+) -> dict[Video, list[Subtitle]]:
     """List subtitles.
 
     The `videos` must pass the `languages` check of :func:`check_video`.
@@ -555,12 +652,12 @@ def list_subtitles(videos, languages, pool_class=ProviderPool, **kwargs):
     :type languages: set of :class:`~babelfish.language.Language`
     :param pool_class: class to use as provider pool.
     :type pool_class: :class:`ProviderPool`, :class:`AsyncProviderPool` or similar
-    :param \*\*kwargs: additional parameters for the provided `pool_class` constructor.
+    :param kwargs: additional parameters for the provided `pool_class` constructor.
     :return: found subtitles per video.
     :rtype: dict of :class:`~subliminal.video.Video` to list of :class:`~subliminal.subtitle.Subtitle`
 
     """
-    listed_subtitles = defaultdict(list)
+    listed_subtitles: dict[Video, list[Subtitle]] = defaultdict(list)
 
     # check videos
     checked_videos = []
@@ -585,14 +682,19 @@ def list_subtitles(videos, languages, pool_class=ProviderPool, **kwargs):
     return listed_subtitles
 
 
-def download_subtitles(subtitles, pool_class=ProviderPool, **kwargs):
+def download_subtitles(
+    subtitles: Sequence[Subtitle],
+    *,
+    pool_class: type[ProviderPool] = ProviderPool,
+    **kwargs: Any,
+) -> None:
     """Download :attr:`~subliminal.subtitle.Subtitle.content` of `subtitles`.
 
     :param subtitles: subtitles to download.
     :type subtitles: list of :class:`~subliminal.subtitle.Subtitle`
     :param pool_class: class to use as provider pool.
     :type pool_class: :class:`ProviderPool`, :class:`AsyncProviderPool` or similar
-    :param \*\*kwargs: additional parameters for the provided `pool_class` constructor.
+    :param kwargs: additional parameters for the provided `pool_class` constructor.
 
     """
     with pool_class(**kwargs) as pool:
@@ -601,8 +703,17 @@ def download_subtitles(subtitles, pool_class=ProviderPool, **kwargs):
             pool.download_subtitle(subtitle)
 
 
-def download_best_subtitles(videos, languages, min_score=0, hearing_impaired=False, only_one=False, compute_score=None,
-                            pool_class=ProviderPool, **kwargs):
+def download_best_subtitles(
+    videos: Set[Video],
+    languages: Set[Language],
+    *,
+    min_score: int = 0,
+    hearing_impaired: bool = False,
+    only_one: bool = False,
+    compute_score: ComputeScore | None = None,
+    pool_class: type[ProviderPool] = ProviderPool,
+    **kwargs: Any,
+) -> dict[Video, list[Subtitle]]:
     """List and download the best matching subtitles.
 
     The `videos` must pass the `languages` and `undefined` (`only_one`) checks of :func:`check_video`.
@@ -618,12 +729,12 @@ def download_best_subtitles(videos, languages, min_score=0, hearing_impaired=Fal
         `hearing_impaired` as keyword argument and returns the score.
     :param pool_class: class to use as provider pool.
     :type pool_class: :class:`ProviderPool`, :class:`AsyncProviderPool` or similar
-    :param \*\*kwargs: additional parameters for the provided `pool_class` constructor.
+    :param kwargs: additional parameters for the provided `pool_class` constructor.
     :return: downloaded subtitles per video.
     :rtype: dict of :class:`~subliminal.video.Video` to list of :class:`~subliminal.subtitle.Subtitle`
 
     """
-    downloaded_subtitles = defaultdict(list)
+    downloaded_subtitles: dict[Video, list[Subtitle]] = defaultdict(list)
 
     # check videos
     checked_videos = []
@@ -641,17 +752,30 @@ def download_best_subtitles(videos, languages, min_score=0, hearing_impaired=Fal
     with pool_class(**kwargs) as pool:
         for video in checked_videos:
             logger.info('Downloading best subtitles for %r', video)
-            subtitles = pool.download_best_subtitles(pool.list_subtitles(video, languages - video.subtitle_languages),
-                                                     video, languages, min_score=min_score,
-                                                     hearing_impaired=hearing_impaired, only_one=only_one,
-                                                     compute_score=compute_score)
+            subtitles = pool.download_best_subtitles(
+                pool.list_subtitles(video, languages - video.subtitle_languages),
+                video,
+                languages,
+                min_score=min_score,
+                hearing_impaired=hearing_impaired,
+                only_one=only_one,
+                compute_score=compute_score,
+            )
             logger.info('Downloaded %d subtitle(s)', len(subtitles))
             downloaded_subtitles[video].extend(subtitles)
 
     return downloaded_subtitles
 
 
-def save_subtitles(video, subtitles, single=False, directory=None, encoding=None):
+def save_subtitles(
+    video: Video,
+    subtitles: Sequence[Subtitle],
+    *,
+    single: bool = False,
+    directory: str | os.PathLike | None = None,
+    encoding: str | None = None,
+    extension: str | None = None,
+) -> list[Subtitle]:
     """Save subtitles on filesystem.
 
     Subtitles are saved in the order of the list. If a subtitle with a language has already been saved, other subtitles
@@ -667,34 +791,35 @@ def save_subtitles(video, subtitles, single=False, directory=None, encoding=None
     :param bool single: save a single subtitle, default is to save one subtitle per language.
     :param str directory: path to directory where to save the subtitles, default is next to the video.
     :param str encoding: encoding in which to save the subtitles, default is to keep original encoding.
+    :param (str | None) extension: the subtitle extension, default is to match to the subtitle format.
     :return: the saved subtitles
     :rtype: list of :class:`~subliminal.subtitle.Subtitle`
 
     """
-    saved_subtitles = []
+    saved_subtitles: list[Subtitle] = []
     for subtitle in subtitles:
         # check content
-        if subtitle.content is None:
+        if not subtitle.content:
             logger.error('Skipping subtitle %r: no content', subtitle)
             continue
 
         # check language
-        if subtitle.language in set(s.language for s in saved_subtitles):
+        if subtitle.language in {s.language for s in saved_subtitles}:
             logger.debug('Skipping subtitle %r: language already saved', subtitle)
             continue
 
         # create subtitle path
-        subtitle_path = get_subtitle_path(video.name, None if single else subtitle.language)
+        subtitle_path = subtitle.get_path(video, single=single, extension=extension)
         if directory is not None:
             subtitle_path = os.path.join(directory, os.path.split(subtitle_path)[1])
 
         # save content as is or in the specified encoding
         logger.info('Saving %r to %r', subtitle, subtitle_path)
         if encoding is None:
-            with io.open(subtitle_path, 'wb') as f:
+            with open(subtitle_path, 'wb') as f:
                 f.write(subtitle.content)
         else:
-            with io.open(subtitle_path, 'w', encoding=encoding) as f:
+            with open(subtitle_path, 'w', encoding=encoding) as f:
                 f.write(subtitle.text)
         saved_subtitles.append(subtitle)
 
